#---------------------------------------------------------------------------+
# Licensed Materials - Property of IBM
# Rational ClearCase
# Copyright IBM Corp. 2005, 2006.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract
# with IBM Corp.
#---------------------------------------------------------------------------+
#
# du_tool.pl
# usage: run with "--help" for usage.
#
# This tool works with Rational Rapid Deploy deployment unit XML files.  It 
# may be invoked to perform one of several operations:
#
# 1. Generation
# 
#    The tool will walk a directory tree to generate a deployment unit XML
#    file with entries for each file found.  
#
# 2. Decoration
#
#    The tool will add VOB object references (OREFs) to files listed in a
#    deployment unit XML file.
#
# 3. Resolution
#
#    The tool will resolve the OREFs in a previously-decorated deployment unit
#    XML file to version-extended pathnames.
#
# 4. Verification
#
#    The tool will parse a deployment unit XML file to verify that it is
#    well-formed XML. It does NOT attempt to validate that the file conforms
#    to the documented XML schema.  (Verification occurs implicitly as part
#    of all other operations.)
#
# Complexities:
#
# All operations (except for verification) require exactly one view context.
# If no view is specified with --tag or --pname, the tool assumes that it is
# already running in a view context and attempts to determine information
# about that view.  
#
# There are two classes of pathname arguments to this tool: 
#     a. Pathnames that must resolve to something within the tool's single
#        view context:
#          --buildlog
#          --directory
#
#     b. Pathnames that may resolve to anything, in or out of the tool's view
#        context.
#          --out
#          input file
#
# Arguments of type (a) must generally be specified with viewroot relative
# paths (e.g. /vobs/my_vob/foo and \my_vob\foo).  Conventional relative paths
# are allowable only when the tool is invoked without an explicitly-named view
# (ie. without --tag or --pname), provided that they are within the same
# implicit view context.  For example, if working somewhere under a dynamic
# view directory (/view/dynamic_view_1/vobs/my_vob),
# "--directory .." is acceptable, but 
# "--directory ../../../dynamic_view_2" is not, since it violates the
# constraint that the tool works in only one view at a time.
#
# Arguments of type (b) are always resolved without regard to a view context.
#
# For those operations that require one, if no input file is specified,
# standard input is assumed.  If no output file is specified, results of all
# operations are printed to standard output.
#

use 5.8.6;
use strict;
use Date::Format;
use Time::Zone;
use Getopt::Long;
use Sys::Hostname;
use XML::Parser;
use XML::Writer;

use ClearCase::RD_utils;

my $orig_dir = RD_getcwd();

set_script_name('du_tool.pl');

use constant USAGE => 
    ClearCase::RD_utils::msgs::USAGE_PREFIX .
    "  du_tool.pl --generate\n" .
    "             --directory staging-dir [[--directory staging-dir]...]\n" .
    "             --name release-name --version release-version\n" .
    "             [--tag dynamic-view-tag | --pname snapshot-view-pname]\n" .
    "             [--out output-file]\n" .
    "\n" .
    "  du_tool.pl { --decorate [--instance instance-id]\n" . 
    "               [[--baseline baseline-text]...]\n" .
    "               [[--logvendor vendor] [--buildlog log-file]...]\n" .
    "             | --resolve\n" .
    "             }\n" .
    "             [--tag dynamic-view-tag | --pname snapshot-view-pname]\n" .
    "             [--out output-file]\n" .
    "             [input-file]\n" .
    "\n" .
    "  du_tool.pl --verify [input-file]\n\n" .
    ClearCase::RD_utils::msgs::USAGE_SUFFIX;

# Mutually-exclusive operations; user must specify exactly one
my $generate_op = undef;
my $decorate_op = undef;
my $resolve_op = undef;
my $verify_op = undef;

# Mutually-exclusive view options; user must specify <= 1
my $view_tag = undef;
my $snapshot_path = undef;

# Generation-specific parameters
my @dirs = ();
my $appname = undef;
my $appver = undef;

# Optional output file argument
my $output_file = undef;

# Optional decoration data
my @baselines = ();
my @buildlogs = ();
my $buildlog_vendor = undef;
my $instance_id = undef;

# Help request
my $help = undef;

if (!GetOptions("help" => \$help,
                "generate" => \$generate_op,
                "decorate" => \$decorate_op,
                "resolve" => \$resolve_op,
                "verify" => \$verify_op,
                "directory=s@" => \@dirs,
                "name=s" => \$appname,
                "version=s" => \$appver,
                "tag=s" => \$view_tag, 
                "pname=s" => \$snapshot_path,
                "out=s" => \$output_file,
                "baseline=s@" => \@baselines,
                "buildlog=s@" => \@buildlogs,
                "logvendor=s" => \$buildlog_vendor,
                "instance=s" => \$instance_id))
{
    pretty_print(ClearCase::RD_utils::MSG_ERR, 
                 ClearCase::RD_utils::msgs::CANT_GET_ARGS . USAGE);
    exit 1;
}

# Print help message and exit immediately if requested
if ($help) {
    pretty_print(ClearCase::RD_utils::MSG_OK, USAGE);
    exit 0;
}

# Ensure that exactly one operation was specified
if (!($generate_op || $decorate_op || $resolve_op || $verify_op) ||
    ($generate_op && $decorate_op) ||
    ($generate_op && $resolve_op) ||
    ($generate_op && $verify_op) ||
    ($decorate_op && $resolve_op) ||
    ($decorate_op && $verify_op) ||
    ($resolve_op && $verify_op))
{
    pretty_print(ClearCase::RD_utils::MSG_ERR, 
                 ClearCase::RD_utils::msgs::ONE_OPERATION . USAGE);
    exit 1;
}

# Handle generation-specific flags
if ($generate_op) {
    if (!@dirs || !$appname || !$appver) {
        pretty_print(ClearCase::RD_utils::MSG_ERR,
                     ClearCase::RD_utils::msgs::GENERATE_REQD_FLAGS . USAGE);
        exit 1;
    }
}
elsif (@dirs || $appname || $appver) {
    pretty_print(ClearCase::RD_utils::MSG_ERR, 
                 ClearCase::RD_utils::msgs::GENERATE_ONLY_FLAGS . USAGE);
    exit 1;
}

# Instance ID, baselines & buildlog stuff only allowed for decoration
if (($instance_id || @baselines || @buildlogs || $buildlog_vendor) && 
    !$decorate_op) 
{
    pretty_print(ClearCase::RD_utils::MSG_ERR,
                 ClearCase::RD_utils::msgs::DECORATE_ONLY_FLAGS . USAGE);
    exit 1;
}

# Can't specify log vendor w/o buildlog(s).
if ($buildlog_vendor && !@buildlogs) {
    pretty_print(ClearCase::RD_utils::MSG_ERR,
                 ClearCase::RD_utils::msgs::VENDOR_REQUIRES_BUILDLOG . USAGE);
    exit 1;
}

# Ensure that view spec makes sense
if ($verify_op && ($view_tag || $snapshot_path)) {
    pretty_print(ClearCase::RD_utils::MSG_ERR,
                 ClearCase::RD_utils::msgs::VIEW_WITH_VERIFY . USAGE);
    exit 1;
}
elsif ($view_tag && $snapshot_path) {
    pretty_print(ClearCase::RD_utils::MSG_ERR, 
                 ClearCase::RD_utils::msgs::ONE_VIEW . USAGE);
    exit 1;
}

# No --out allowed with --verify
if ($verify_op && $output_file) {
    pretty_print(ClearCase::RD_utils::MSG_ERR, 
                 ClearCase::RD_utils::msgs::OUT_WITH_VERIFY . USAGE);
    exit 1;
}

# If specified, input file is the only remaining argument - but isn't used for
# generation.
my $input_file = ($generate_op ? undef : shift(@ARGV));

# Ensure no additional arguments
if (@ARGV) {
    foreach my $arg (@ARGV) {
        pretty_print(ClearCase::RD_utils::MSG_ERR,
                     ClearCase::RD_utils::msgs::INVALID_OPTION, 
                     $arg);
    }
    pretty_print(ClearCase::RD_utils::MSG_OK, USAGE);
    exit 1;
}

# File/path argument cleanup; paths may have been quoted & wrapped in
# whitespace.  This will cause all sorts of ugliness, so remove them now.
foreach my $quoted_path (@dirs, $snapshot_path, $output_file, @buildlogs) {
    next unless $quoted_path;
    $quoted_path =~ s/\"//;
    remove_external_whitespace($quoted_path);
}

# Set up input stream; ensure that input is treated as UTF-8
my $input_fh = undef;
if (!$generate_op) {
    if (!$input_file) {
        set_binmode(ClearCase::RD_utils::RD_STDIN, ':utf8');
        $input_fh = \*STDIN;
    }
    elsif (!open($input_fh, '<:utf8', $input_file)) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::CANT_OPEN, 
                     $input_file, 
                     "$!");
        exit 1;
    }
}

# Set up output stream; file must not exist yet, and output is UTF-8
my $output_fh = undef;
if (!$output_file) {
    set_binmode(ClearCase::RD_utils::RD_STDOUT, ':utf8');
    $output_fh = \*STDOUT;
}
else {
    if (file_test(ClearCase::RD_utils::FILE_TEST_EXISTS, $output_file)) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::FILE_EXISTS, 
                     $output_file);
        exit 1;
    }
    
    if (!open($output_fh, '>:utf8', $output_file)) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::CANT_OPEN, 
                     $output_file, 
                     "$!");
        exit 1;
    }
}

my $started_view = undef;
my $view_dir = undef;
my $verification_view_tag = undef;

if (!$verify_op) {
    my $cmd_result = undef;

    # Set up the view directory
    if ($view_tag) {
        # Only start the view if necessary
        (my $seems_a_view, $view_dir) = dynamic_view_is_started($view_tag);
        
        # Error case; function already blurted.
        exit 1 unless defined($seems_a_view);

        if (!$seems_a_view && 
            do_cmd("$cleartool startview $view_tag", $cmd_result))
        {
            $seems_a_view = 1;
            $started_view = 1;
        }

        if (!$seems_a_view || !dynamic_view_is_started($view_tag)) {
            pretty_print(ClearCase::RD_utils::MSG_ERR,
                         ClearCase::RD_utils::msgs::NOT_DYNAMIC_VIEW,
                         $view_tag);
            exit 1;
        }

        $verification_view_tag = $view_tag;
    }
    elsif ($snapshot_path) {
        ($verification_view_tag, $view_dir, my $view_pname) = 
            get_view_info_from_pname($snapshot_path);

        unless ($verification_view_tag) {
            pretty_print(ClearCase::RD_utils::MSG_ERR,
                         ClearCase::RD_utils::msgs::NOT_SNAPSHOT_VIEW,
                         $snapshot_path);
            exit 1;
        }
    }
    else {
        # First ensure that we're in a view context.
        $verification_view_tag = get_cwv_tag();

        if (!$verification_view_tag) {
            # Function already blurted if return value is undef.
            if (defined($verification_view_tag)) {
                pretty_print(ClearCase::RD_utils::MSG_ERR, 
                             ClearCase::RD_utils::msgs::NEED_VIEW_CTX);
            }
            exit 1;
        }

        # Assume for now that the tag belongs to a dynamic view.  If so, it
        # MUST have been started since we just ensured that we're in it!
        (my $already_started, $view_dir) = 
            dynamic_view_is_started($verification_view_tag);
        
        # Error case; function already blurted.
        exit 1 unless defined($already_started);

        # If not started, it must be a snapshot view.
        unless ($already_started) {
            ($verification_view_tag, $view_dir, my $view_pname) = 
                get_view_info_from_pname($orig_dir);
        }
    }

    exit 1 if (!$view_dir);
}

# Now that we've ensured a view context, we need to process any of the files
# described in the help notes above as type (a), by:
#     - converting relative paths to absolute paths within the working view
#       (if working in an implicit view)
#     - ensuring that those absolute paths don't point off to some other view
if ($view_dir) {
    my $allow_rel_paths = !$view_tag && !$snapshot_path;
    my $num_bad_paths = 0;
    my $path = undef;

    # Convert relative to absolute, if appropriate.
    foreach $path (@dirs, @buildlogs) {
        next if file_name_is_absolute($path);

        $num_bad_paths++;

        if ($allow_rel_paths) {
            $path = canonpath(rel2abs($path));

            my ($path_view_tag, $path_view_dir, $new_path) = 
                get_view_info_from_pname($path);

            # This will catch errors with the function call AND complaints if
            # $path is not in any view at all.
            next unless $path_view_tag;

            # Different view?
            if ($path_view_tag ne $verification_view_tag) {
                pretty_print(ClearCase::RD_utils::MSG_ERR,
                             ClearCase::RD_utils::msgs::NO_MULTIPLE_VIEW_PATHS,
                             $path,
                             $verification_view_tag);
                next;
            }
            $path = $new_path;
        }
        else {
            pretty_print(ClearCase::RD_utils::MSG_ERR,
                         ClearCase::RD_utils::msgs::NO_RELATIVE_PATHS,
                         $path);
            next;
        }

        $num_bad_paths--;
    }

    exit 1 if ($num_bad_paths > 0);

    # Move to the view context.
    unless (RD_chdir($view_dir)) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::DIR_ACCESS, 
                     $view_dir, 
                     "$!");
        exit 1;
    }
}

use constant XML_INDENT => 4;

# Don't need a source if we're generating.  Don't need a target if we're
# verifing.
my $xml_source = 
    ($generate_op ? 
     undef :
     new XML::Parser(Style => 'Stream', 
                     Handlers => { XMLDecl => \&handle_xml_decl,
                                   Comment => \&handle_xml_comment}));
my $xml_target = undef;
my $curr_output_step = 0;
my @history_elems = (); # Contains attr hash refs for history elements
my $ts_time = undef;
my $num_file_elems = 0;
my $within_elided_element = 0;

if (!$verify_op) {
    $xml_target = new XML::Writer(OUTPUT => $output_fh,
                                  DATA_MODE => 1,
                                  DATA_INDENT => XML_INDENT);

    # Need an XML declaration before anything else.  (We can't pass through
    # from any input doc because it's consumed by the $xml_source object.)
    $xml_target->xmlDecl('UTF-8');

    # Save the current time; it will be useful when decorating, when we'll
    # need to express it in a different format than we use for history
    # elements.
    $ts_time = time();

    my $timestamp = do_time2str($ts_time);

    my $localhost = undef;

    # Function call may fail, but it will croak on its own.
    if ($localhost = hostname()){
        # Expand to fully-qualified hostname, if possible.
        my ($tmp_host, $aliases) = gethostbyname($localhost);
        ($localhost) = split(/\s+/, $aliases) if ($aliases);
    }
    else {
        $localhost = '?';
    }
    $localhost = native_cp_to_internal_utf8($localhost) or exit 1;
    
    # Compute user name; see getlogin() documentation in "Programming
    # Perl, 3rd Edition" by Wall, Christiansen & Orwant

    my $user = getlogin() || (getpwuid($<))[0];
    if (!$user) {
	pretty_print(ClearCase::RD_utils::MSG_WARN, 
                     ClearCase::RD_utils::msgs::NO_USER);
	$user = '?';
    }
    $user = native_cp_to_internal_utf8($user) or exit 1;

    $history_elems[scalar(@history_elems)] = 
        {ClearCase::RD_utils::HISTORY_OPERATION_ATTR() => 
             ($generate_op ? ClearCase::RD_utils::HISTORY_OP_GENERATE : 
              ($decorate_op ? ClearCase::RD_utils::HISTORY_OP_DECORATE : 
               ClearCase::RD_utils::HISTORY_OP_RESOLVE)),
              ClearCase::RD_utils::HISTORY_HOST_ATTR() => $localhost,
              ClearCase::RD_utils::HISTORY_USER_ATTR() => $user,
              ClearCase::RD_utils::HISTORY_TS_ATTR() => $timestamp,
              ClearCase::RD_utils::HISTORY_NOTE_ATTR() => ''};
}

# XML::Parser and XML::Writer die if they encounter errors.
eval {
    if ($generate_op) {
        (($appname = native_cp_to_internal_utf8($appname)) &&
         ($appver = native_cp_to_internal_utf8($appver)) ||
         exit 1);

        $xml_target->startTag(ClearCase::RD_utils::XML_DOC_ELEMENT,
                              ClearCase::RD_utils::XML_XSI_ATTR() => 
                              ClearCase::RD_utils::XML_XSI_VAL,
                              ClearCase::RD_utils::XML_TRACKING_ATTR() => 
                              ClearCase::RD_utils::XML_TRACKING_VAL,
                              ClearCase::RD_utils::XML_APPNAME_ATTR() => 
                              $appname,
                              ClearCase::RD_utils::XML_APPVER_ATTR() => 
                              $appver);

        my $total_num_files_written = 0;

        foreach my $curr_dir (@dirs) {
            exit 1 if (!write_dir_contents($view_dir, 
                                           $curr_dir, 
                                           undef,
                                           \$total_num_files_written));
        }

        if (!$total_num_files_written) {
            pretty_print(ClearCase::RD_utils::MSG_WARN, 
                         ClearCase::RD_utils::msgs::FOUND_NO_FILES);
            exit 1;
        }

        $xml_target->emptyTag(ClearCase::RD_utils::XML_HISTORY_ELEMENT, 
                              %{$history_elems[0]});

        $xml_target->endTag(ClearCase::RD_utils::XML_DOC_ELEMENT);
    }
    else {
        $xml_source->parse($input_fh) if ($xml_source);
        if (!$verify_op && $num_file_elems == 0) {
            pretty_print(ClearCase::RD_utils::MSG_WARN, 
                         ClearCase::RD_utils::msgs::NO_FILE_ELEMENTS);
            exit 1;
        }
    }

    $xml_target->end() if $xml_target;
};
if ($@) {
    # Generation code already blurted.
    if (!$generate_op) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_FILE, 
                     "$@");
    }
    exit 1;
}

pretty_print(ClearCase::RD_utils::MSG_OK, 
             ClearCase::RD_utils::msgs::VALID_FILE) if ($verify_op);

END {
    # Save exit status in case any of our cleanup work modifies it
    my $exit_status = $?;

    # Close input/output filehandles only if specified as actual files!
    close($input_fh) if ($input_file && $input_fh);

    # Revert to original working directory.  We didn't canonicalize the name
    # of the output file, so deleting could cause problems.  (We also may not
    # be able to end a dynamic view safely if we're still within one of its
    # directories.
    my $safe_to_delete = ($orig_dir eq RD_getcwd());

    if ($orig_dir && !$safe_to_delete) {
        if (RD_chdir($orig_dir)) {
            $safe_to_delete = 1;
        }
        else {
            pretty_print(ClearCase::RD_utils::MSG_WARN, 
                         ClearCase::RD_utils::msgs::DIR_ACCESS, 
                         $orig_dir,
                         "$!");
        }
    }

    if ($output_file && $output_fh) {
        if (!close($$output_fh)) {
            pretty_print(ClearCase::RD_utils::MSG_WARN, 
                         ClearCase::RD_utils::msgs::CANT_CLOSE, 
                         $output_file, 
                         "$!");
        }
        
        # Delete the output file if we didn't populate it completely
        if (($exit_status != 0) && $safe_to_delete) {
            pretty_print(ClearCase::RD_utils::MSG_WARN, 
                         ClearCase::RD_utils::msgs::REMOVING_BAD_OUTPUT,
                         $output_file);
            
            if (!unlink($output_file)) {
                pretty_print(ClearCase::RD_utils::MSG_WARN, 
                             ClearCase::RD_utils::msgs::CANT_UNLINK, 
                             $output_file, 
                             "$!");
            }
        }
    }
    
    # End the dynamic view that we started
    if ($started_view) {
        my $cmd_result = undef;
        do_cmd("$cleartool endview $view_tag", $cmd_result);
    }
    
    # Exit with non-zero status if errors occurred.  Return the earliest
    # possible error.
    exit ($exit_status == 0 ? $? : $exit_status);

} # END


################################################################################
# handle_xml_decl -       Ensure (by dying) that input document conforms to
#                         supported XML version and encoding.
#          
# IN   expat              Underlying expat instance
# IN   version            XML version string
# IN   encoding           Character encoding
# IN.  standalone         Standalone flag
sub handle_xml_decl($$$$)
{
    my ($expat, $version, $encoding, $standalone) = @_;

    if ($version ne "1.0") {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_VERSION, 
                     $version);
        die;
    }

    my $lc_encoding = lc($encoding);

    # We support ASCII and UTF-8 only.  (Convenient side-effect of this choice
    # is that we can unconditionally assume UTF-8 when decoding, since ASCII
    # is a proper subset.)  If no encoding is defined, we assume UTF-8.
    if ($lc_encoding && 
        ($lc_encoding ne "utf-8") && 
        ($lc_encoding ne "us-ascii")) 
    {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_ENCODING, 
                     $encoding);
        die;
    }

} # handle_xml_decl 


################################################################################
# handle_xml_comment -    Handle XML comments found while parsing an input
#                         file.
#          
# IN   expat              Underlying expat instance
# IN   data               Comment data
sub handle_xml_comment($$) 
{
    my ($expat, $data) = @_;

    return if ($verify_op || $within_elided_element);
    remove_external_whitespace($data);
    $xml_target->comment($data);
} # handle_xml_comment


################################################################################
# StartTag -              Subroutine called by XML::Parser object when it
#                         encounters a start tag.  The complete tag is stored
#                         in $_, while the attributes are hashed in %_.
#          
# IN   expat              Underlying expat instance.
# IN   eltype             Element type (the tag's name).
sub StartTag($$)
{
    my ($expat, $eltype) = @_;
    my %attrs = %_;

    # Elide any elements that may already exist in the file that we'd
    # ordinarily add during decoration; this allows us to redecorate an
    # already-decorated file and still conform to the DU XML schema.
    return if ($verify_op || $within_elided_element);

    if ($decorate_op && 
        ClearCase::RD_utils::DECORATE_CREATED_ELEMENTS->{$eltype}) 
    {
        $within_elided_element = 1;
        return;
    }

    # Document, file and history elements need special handling; all other
    # elements should be passed through unmolested.
    if ($eltype eq ClearCase::RD_utils::XML_DOC_ELEMENT) {
        # If decorating, we've got to add/update the instance ID before
        # emitting the document element.
        if ($decorate_op) {
            # Need to use some portion of the current time to help
            # disambiguate instances that are created on the same day.  Let's
            # be paranoid and use the entire time.
            my $timestamp = do_time2str($ts_time, 1);

            # Use the user-specified instance ID as prefix; otherwise, use the
            # (possibly non-existent) release name.
            my $instance_attr_prefix = $instance_id;
            my $release_name = $attrs{ClearCase::RD_utils::XML_APPNAME_ATTR()};

            unless ($instance_attr_prefix && $release_name) {
                $instance_attr_prefix = 
                    internal_utf8_to_native_cp($release_name) or die;
            }

            # Strip leading/trailing whitespace
            remove_external_whitespace($instance_attr_prefix);
            
            # Collapse whitespace down to underscores.
            $instance_attr_prefix =~ s/\s+/_/g;

            $instance_attr_prefix = 
                native_cp_to_internal_utf8($instance_attr_prefix) or die;

            # Don't need to add any separation string between the prefix and
            # the timestamp if the prefix is empty.
            $attrs{ClearCase::RD_utils::XML_INSTANCE_ATTR()} = 
                $instance_attr_prefix . 
                ($instance_attr_prefix ? '-' : '') .
                $timestamp;
        }
        
        $xml_target->startTag($eltype, %attrs);

    }
    elsif ($eltype eq ClearCase::RD_utils::XML_FILE_ELEMENT) {
        if ($decorate_op) {
            my $pname = $attrs{ClearCase::RD_utils::FILE_PNAME_ATTR()};

            if (!$pname) {
                pretty_print(ClearCase::RD_utils::MSG_ERR, 
                             ClearCase::RD_utils::msgs::MISSING_ATTR, 
                             ClearCase::RD_utils::FILE_PNAME_ATTR,
                             $eltype,
                             $expat->current_line());
                die;
            }

            $pname = internal_utf8_to_native_cp($pname) or die;

            my ($oref_attr, $cs_rule_attr) = 
                convert_pname_to_uri($pname, $view_dir);
            
            if (!$oref_attr || !$cs_rule_attr) {
                pretty_print(ClearCase::RD_utils::MSG_ERR,
                             ClearCase::RD_utils::msgs::CANT_CONVERT_PNAME,
                             $pname,
                             $expat->current_line());
                die;
            }

            $attrs{ClearCase::RD_utils::FILE_OREF_ATTR()} = $oref_attr;
            $attrs{ClearCase::RD_utils::FILE_CSRULE_ATTR()} = $cs_rule_attr;
            delete($attrs{ClearCase::RD_utils::FILE_XNAME_ATTR()});
        }
        else {
            my $oref_attr = $attrs{ClearCase::RD_utils::FILE_OREF_ATTR()};
   
            if (!$oref_attr) {
                pretty_print(ClearCase::RD_utils::MSG_ERR, 
                             ClearCase::RD_utils::msgs::MISSING_ATTR, 
                             ClearCase::RD_utils::FILE_OREF_ATTR,
                             $eltype,
                             $expat->current_line());
                die;
            }

            $oref_attr = internal_utf8_to_native_cp($oref_attr) or die;

            $attrs{ClearCase::RD_utils::FILE_XNAME_ATTR()} = 
                native_cp_to_internal_utf8(convert_uri_to_pname($oref_attr)) 
                or die;
        }

        # These should always be emitted as empty elements.
        $xml_target->emptyTag($eltype, %attrs);
        $num_file_elems++;
    }
    elsif ($eltype eq ClearCase::RD_utils::XML_HISTORY_ELEMENT) {
        # Save this empty element for later processing; we want to spit these
        # out all at once at the end of the document.
        $history_elems[scalar(@history_elems)] = \%attrs;
    }
    elsif (ClearCase::RD_utils::EMPTY_XML_ELEMENTS->{$eltype}) {
        $xml_target->emptyTag($eltype, %attrs);
    }
    else {
        $xml_target->startTag($eltype, %attrs);
    }
} # StartTag


################################################################################
# compute_indent_string - Compute and return whitespace to indent XML output.
#          
# IN   depth              Current depth within the XML document
# RESULT:                 Indent string
sub compute_indent_string($) 
{
    my ($depth) = @_;
    
    return ($depth <= 0) ? "" : (" " x XML_INDENT) x $depth;
} # compute_indent_string


################################################################################
# EndTag -                Subroutine called by XML::Parser object when it
#                         encounters an end tag.  The complete tag is stored
#                         in $_.
#                         
# IN   expat              Underlying expat instance.
# IN   eltype             Element type (the tag's name).
sub EndTag($$) 
{
    my ($expat, $eltype) = @_;

    # If it appears in EMPTY_XML_ELEMENTS, we've already emitted an empty tag.
    # If we're decorating, we've chosen to elide some elements in StartTag.
    return if ($verify_op || 
               ClearCase::RD_utils::EMPTY_XML_ELEMENTS->{$eltype});

    if ($decorate_op && 
        ClearCase::RD_utils::DECORATE_CREATED_ELEMENTS->{$eltype}) 
    {
        $within_elided_element = 0;
        return;
    }

    return if ($within_elided_element);

    if ($eltype eq ClearCase::RD_utils::XML_DOC_ELEMENT) {
        # If we're looking at the document element and are decorating, we need
        # to emit the origin and buildlog elements.
        if ($decorate_op) {

            # Need the config spec and any user-supplied baseline info in an
            # origin element.
            $xml_target->startTag(ClearCase::RD_utils::XML_ORIGIN_ELEMENT);

            foreach my $bl (@baselines) {
                # Elide empty baselines.
                next unless $bl;

                $bl = native_cp_to_internal_utf8($bl) or die;
                $xml_target->emptyTag(ClearCase::RD_utils::XML_BASELINE_ELEMENT,
                                      ClearCase::RD_utils::BASELINE_NAME_ATTR() 
                                      => $bl,
                                      ClearCase::RD_utils::BASELINE_AUTH_ATTR()
                                      => 
                                      ClearCase::RD_utils::BASELINE_AUTH_VAL);
            }

            my $config_spec = undef;

            die unless do_cmd("$cleartool catcs", $config_spec);

            # Need to pretty up the config spec lines to maintain readbility.
            my $curr_depth = $expat->depth() + 3;
            my $data_indent_str = compute_indent_string($curr_depth);
            my @lines = split(/^/, $config_spec);

            $config_spec = "\n";

            foreach my $line (@lines) {
                remove_external_whitespace($line);
                next if (!$line);
                $config_spec .= $data_indent_str . $line . "\n";
            }

            $config_spec .= compute_indent_string($curr_depth - 1);
            $config_spec = native_cp_to_internal_utf8($config_spec) or die;

            $xml_target->dataElement(ClearCase::RD_utils::XML_CS_ELEMENT,
                                     $config_spec);
            $xml_target->endTag(ClearCase::RD_utils::XML_ORIGIN_ELEMENT);

            if (@buildlogs) {
                $xml_target->startTag(ClearCase::RD_utils::XML_BLDORGN_ELEMENT);

                my $num_bad_logs = 0;
                my %log_attrs = ();

                if ($buildlog_vendor) {
                    $log_attrs{ClearCase::RD_utils::BLDLG_VENDOR_ATTR()} = 
                        native_cp_to_internal_utf8($buildlog_vendor) or die;
                }

                foreach my $build_log_path (@buildlogs) {
                    # Assume that we have a problem with this build log until
                    # we complete the loop iteration.  (It'll be friendliest
                    # to complain about ALL unresolvable/unconvertable build
                    # logs.)
                    $num_bad_logs++;

                    my $uri = convert_pname_to_uri($build_log_path, $view_dir) 
                        or next;

                    $log_attrs{ClearCase::RD_utils::FILE_PNAME_ATTR()} = 
                        native_cp_to_internal_utf8($build_log_path) or next;
                    
                    $log_attrs{ClearCase::RD_utils::FILE_OREF_ATTR()} = $uri;

                    $xml_target->emptyTag(ClearCase::RD_utils::XML_BLDLG_ELEM, 
                                          %log_attrs);
                    $num_bad_logs--;
                }

                die if ($num_bad_logs > 0);
                $xml_target->endTag(ClearCase::RD_utils::XML_BLDORGN_ELEMENT);

            } # any build logs to emit?

        } # decorating?

        # Accumulated history elements are always the last thing we emit in
        # the document.
        foreach my $hist_attrs (@history_elems) {
            $xml_target->emptyTag(ClearCase::RD_utils::XML_HISTORY_ELEMENT, 
                                  %{$hist_attrs});
        }
    } 

    # Pass through to output XML document.
    $xml_target->endTag($eltype);

} # EndTag


################################################################################
# Text -                  Subroutine called by XML::Parser object when it
#                         encounters character data outside of a tag.  The
#                         data is stored in $_.
#                         
# IN   expat              Underlying expat instance.
sub Text()
{
    # Pass through to output XML document.  Don't bother to emit
    # whitespace-only data.
    return if ($verify_op || $within_elided_element || !($_ =~ /\S/));
    $xml_target->characters($_);
} # Text


################################################################################
# write_dir_contents -    Write XML file elements for the directory contents
#                         of the specified path.  The path is the
#                         concatenation of the two specified parameters
#                         view_path and vob_path.  The path attribute of the
#                         file element will only include the VOB path and not
#                         the system absolute path.
#          
# IN   view_path          The directory path of a ClearCase view.
# IN   vob_path           The VOB path within the ClearCase view.
# IN   orig_path_root     Path to be prepended to files found in the vob_path
#                         directory in file element's orig_path attribute.
#                         (Caller should initialize to undef/empty string, as
#                         this is really only for recursive use.)
# IO   num_files_written  Total number of files written by this recursive
#                         routine.  Caller must initialize to 0 for accurate
#                         results.
# RESULT:                 True if successful.  (Blurts and returns undef if
#                         errors occur.  May die if XML output failure occur.)
sub write_dir_contents($$$;\$)
{
    my ($view_path, $vob_path, $orig_path_root, $num_files_written) = @_;

    if (!$view_path || !$vob_path) {
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL,
                     "write_dir_contents");
        return undef;
    }

    # get the directory contents
    my $dirpath = catdir($view_path, $vob_path);

    if (!file_test(ClearCase::RD_utils::FILE_TEST_DIRECTORY, $dirpath)) {
        # Print the view-root-relative path since that's what the user
        # originally specified.
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::NOT_A_DIRECTORY, 
                     $vob_path);
        return undef;
    }

    if (!opendir DIR_HANDLE, $dirpath) {
        # Use the concatenated path here since we can't make any mention with
        # DIR_ACCESS of "the specified view"; full path should help with user
        # diagnosis.
        pretty_print(ClearCase::RD_utils::MSG_ERR, 
                     ClearCase::RD_utils::msgs::DIR_ACCESS, 
                     $dirpath, 
                     "$!");
        return undef;
    }

    my @contents = grep !/^\.\.?$/, readdir DIR_HANDLE;
    closedir DIR_HANDLE;

    my @dir_list = ();

    # first pass process all file entries in the directory
    foreach my $entry (@contents) {
        # compose the VOB path and full path of the entry
        my $entry_vob_path = catfile($vob_path, $entry);
        my $entry_path = cat_view_dir_and_pname($entry_vob_path, $view_path);
                                                
        if (file_test(ClearCase::RD_utils::FILE_TEST_DIRECTORY, $entry_path)) {
            # for directory, save it to the list and process them later
            push(@dir_list, [$entry_vob_path, $entry]);
            next;
        } 

        # We can't do any useful cleartool operations on these files without
        # chasing links first.  However, it's important to attempt to preserve
        # whatever meaning exists in the layout of the directory passed to
        # this function, including the original filenames.  Therefore, we
        # do CC operations on chased/resolved pathnames, but emit the original
        # names.
        my $resolved_vob_path = chase_symlinks($entry_vob_path, $view_path) 
            or return undef;
        my $resolved_path = cat_view_dir_and_pname($resolved_vob_path, 
                                                   $view_path);

        # Assume that failure here means some dangling symlink or other issue
        # that should NOT otherwise stop our directory scan.  This call will
        # blurt on its own.
        my $meta_type = get_object_meta_type($resolved_path);
        
        next unless $meta_type;

        unless ($meta_type == ClearCase::RD_utils::VERSION_MT) {
            my $msg = ($meta_type == ClearCase::RD_utils::VIEW_PVT_MT ?
                       ClearCase::RD_utils::msgs::SKIPPING_VIEW_PVT_OBJ :
                       ($meta_type == ClearCase::RD_utils::NULL_MT ?
                        ClearCase::RD_utils::msgs::SKIPPING_NULL_META_TYPE :
                        ClearCase::RD_utils::msgs::SKIPPING_NON_VERSION_TYPE));

            pretty_print(ClearCase::RD_utils::MSG_WARN, $msg, $entry_vob_path);
            next;
        }

        my $orig_path = 
            ($orig_path_root ? catfile($orig_path_root, $entry) : $entry);

        (($entry_vob_path = native_cp_to_internal_utf8($entry_vob_path)) &&
         ($orig_path = native_cp_to_internal_utf8($orig_path))) 
            or return undef;

        # for file, write out XML file element with the VOB path
        $xml_target->emptyTag(ClearCase::RD_utils::XML_FILE_ELEMENT, 
                              ClearCase::RD_utils::FILE_PNAME_ATTR() => 
                              $entry_vob_path,
                              ClearCase::RD_utils::FILE_ORIG_ATTR() => 
                              $orig_path);
        
        ($$num_files_written)++ if $num_files_written;
    }

    # process saved directory entries if there are any
    foreach my $entry_pair (@dir_list) {
        my $num_files_this_dir = 0;

        # write out the directory contents recursively
        write_dir_contents($view_path, 
                           $entry_pair->[0],
                           ($orig_path_root ? 
                            catdir($orig_path_root, $entry_pair->[1]) :
                            $entry_pair->[1]),
                           \$num_files_this_dir) or return undef;

        ($$num_files_written) += $num_files_this_dir if $num_files_written;
    }

    return 1;

} # write_dir_contents


my $tz_offset = undef;
my $tz_offset_dense = undef;


################################################################################
# do_time2str -           Stringify a time value.
#          
# IN   timeval            Time value to stringify.
# IN   do_dense           Produce a densely-packed string, or include
#                         punctuation?
# RESULT:                 Formatted string.
sub do_time2str($;$)
{
    my ($timeval, $do_dense) = @_;

    # ISO8601 and dense (punctuationless) formats:
    # ("2005-12-31T14:30:56-04:00")
    # ("20051231143056-0400")
    use constant TIME_FMT => "%Y-%m-%dT%H:%M:%S";
    use constant DENSE_TIME_FMT => "%Y%m%d%H%M%S";

    my $output = time2str(($do_dense ? DENSE_TIME_FMT : TIME_FMT), $timeval);

    # time2str's %z gives us the timezone offset in hours, but we need hh:mm.
    # No need to calculate it more than once, though.
    unless ($tz_offset && $tz_offset_dense) {
        use integer; # Gives us integer division

        my $full_offset_mins = tz_local_offset() / 60;
        my $tz_offset_sign = ($full_offset_mins < 0 ? '-' : '+');

        $full_offset_mins = abs($full_offset_mins);
        
        my $fmt = '%02u';
        my $tz_offset_hrs = sprintf($fmt, $full_offset_mins / 60);
        my $tz_offset_mins = sprintf($fmt, $full_offset_mins % 60);
        my $tz_offset_str_common = $tz_offset_sign . $tz_offset_hrs;

        $tz_offset = $tz_offset_str_common . ':' . $tz_offset_mins;
        $tz_offset_dense = $tz_offset_str_common . $tz_offset_mins;
    }
    
    $output .= ($do_dense ? $tz_offset_dense : $tz_offset);

    # Some of the format specifiers may produce spaces instead of leading
    # zeroes, but we don't want spaces.
    $output =~ s/ /0/g;
    
    return $output;

} # do_time2str
